/**************************************************************
 * File: agrsoftmodem.c
 *
 * Copyright (c) 2002, 2003, 2004, 2005 Agere Systems, Inc.  All rights reserved.
 *
 * Description:
 *   Agere Soft Modem driver interface functions for Linux
 *
 * Revision History:
 *   Name                   Date          Change
 *   Soumyendu Sarkar       12/03/2002    Initial as agrmodem.c
 *   Soumyendu Sarkar       03/11/2004    Adapted to kernel 2.6.x with PCI support
 *   Soumyendu Sarkar       11/16/2005    Adapted to kernel 2.6.14 with PCI PP modem support
 *************************************************************/
 
#define LUCENT_MODEM
#ifndef AGERE_SOFT_MODEM
#define AGERE_SOFT_MODEM
#endif

#include <linux/config.h>
#include <linux/module.h>

#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>

#include <linux/sched.h>

#include <linux/ctype.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/time.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/smp.h>
#include <asm/io.h>
#include <asm/atomic.h>
#include "linuxif.h"
#include "agrmodem.h"

extern unsigned int BaseAddress;

#ifndef LT_KER_26
#define LT_KER_26 1
#endif

int numSupportedDevice;
/** list of supported devices - vendorID, deviceID, subSystemID, subVendorID **/
/* if both subSystemID and subVendorID are set to 0, the driver will ignore these fields */
const SUPPORTED_DEVICES agereSupDevices[] = {
  { 0x11C1, 0x0620, 0x0000, 0x0000 },
  { 0x8086, 0x2416, 0x0000, 0x0000 },
  { 0x8086, 0x2426, 0x0000, 0x0000 },
  { 0x8086, 0x2446, 0x0000, 0x0000 },
  { 0x8086, 0x7196, 0x0000, 0x0000 },
  { 0x8086, 0x2486, 0x0000, 0x0000 },
  { 0x8086, 0x24C6, 0x0000, 0x0000 },
  { 0x8086, 0x24D6, 0x0000, 0x0000 },
  { 0x11C1, 0x048C, 0x0000, 0x0000 },
  { 0x11C1, 0x048F, 0x0000, 0x0000 },
  { 0x8086, 0x266D, 0x0000, 0x0000 },
};

/************** Ensure Kernel Version 2.4.0 or later *************/
#if LINUX_VERSION_CODE < (((2)<<16)|((4)<<8)|((0)))
#error This driver compiles for Linux versions 2.4.0 and later
#endif

static char *modem_name = "Agere Modem Controller driver";
static char *modem_version = "2.1.60.0.Linspire";
static char *modem_revdate = "2005-11-16";
extern void x_debug_msg (char *);
struct timer_list timerList;
int (*agr_rs_interrupt)(void) = NULL;
spinlock_t modem_driver_lock = SPIN_LOCK_UNLOCKED;

/********************* Interface Operations *********************/
static int GetAgrModemInterface(void *mdmdata)
{
	struct ltmodem_ops *lucent_modem_ops;

	lucent_modem_ops = (struct ltmodem_ops *)mdmdata;
	lucent_modem_ops->detect_modem = wrap_lucent_detect_modem;
	lucent_modem_ops->init_modem = wrap_lucent_init_modem;
	lucent_modem_ops->PortOpen = modemPortOpen;
	lucent_modem_ops->PortClose = modemPortClose;
	lucent_modem_ops->read_vuart_register = wrap_read_vuart_port;
	lucent_modem_ops->write_vuart_register = wrap_write_vuart_port;
	lucent_modem_ops->app_ioctl_handler = wrap_app_ioctl_handler;
	lucent_modem_ops->dsp_isr = wrap_dum_dp_dsp_isr;
	lucent_modem_ops->read_buffer = wrap_V16550_Read_RBR_buffer;
  return TRUE;
}

static int SetAgrModemInterface(int state)
{
	if (state == 1)
		agr_rs_interrupt = inter_module_get_request("agr_rs_interrupt", "");
	else
	{
		inter_module_put("agr_rs_interrupt");
		agr_rs_interrupt = NULL;
	}
	if (agr_rs_interrupt == NULL)
		return (-1);
	return (0);
}

/************************ Open / Close *************************/
int modemPortOpen ( void )
{
	wrap_linux_modem_open ();
	return( wrap_vxdPortOpen() );
}

void modemPortClose ( void )
{
	wrap_vxdPortClose();
	SetAgrModemInterface(0);
	wrap_linux_modem_close();
}

/************************** Entry Points ***************************/
int __init modem_init_module (void)
	{
  int ret = 0;
  numSupportedDevice = sizeof(agereSupDevices) / sizeof(agereSupDevices[0]);
  ret = wrap_linux_modem_init ();
  if (!ret) {
	  printk("<1>Loading module %s version %s (%s)\n", modem_name, modem_version, modem_revdate);
	  inter_module_register("GetAgrModemInterface", THIS_MODULE, GetAgrModemInterface);
	  inter_module_register("SetAgrModemInterface", THIS_MODULE, SetAgrModemInterface);
  }
  else
	  printk("<1>Could not detect Agere soft modem device\n Agere soft modem driver not loaded\n");

  return (ret);
	}

void __exit modem_cleanup_module (void)
	{
	wrap_linux_modem_cleanup ();
	inter_module_unregister("SetAgrModemInterface");
	inter_module_unregister("GetAgrModemInterface");
	printk("<1>Unloading %s: version %s\n", modem_name, modem_version);
	}

module_init(modem_init_module);
module_exit(modem_cleanup_module);
MODULE_AUTHOR("Agere Systems Inc");
MODULE_DESCRIPTION("Agere Modem Controller driver");
MODULE_LICENSE("Proprietary");

/************************ Memory allocation ************************/
fnatr void *x_vmalloc (unsigned int size) { return vmalloc (size); }
fnatr void x_vfree (void *ptr) { vfree (ptr); }
fnatr void *x_kmalloc (unsigned int size) { return kmalloc (size, GFP_KERNEL | GFP_DMA); }
fnatr void x_kfree (void *ptr) { kfree (ptr); }

fnatr unsigned x_virt_to_bus (void *virt_addr) { return virt_to_bus (virt_addr); }
fnatr void *x_ioremap_nocache (unsigned int phys_addr, unsigned size) { return ioremap_nocache (phys_addr, size); }
fnatr void x_iounmap (void *virt_addr) { iounmap (virt_addr); }
	
/************************* I/O Functions ***************************/
fnatr unsigned char agr_inp (unsigned address) { return inb (address); }
fnatr unsigned short agr_inpw (unsigned address) { return inw (address); }
fnatr unsigned agr_inpl (unsigned address) { return inl (address); }

fnatr void agr_outp (unsigned address, unsigned char value) { outb (value, address); }
fnatr void agr_outpw (unsigned address, unsigned short value) { outw (value, address); }
fnatr void agr_outpl (unsigned address, unsigned value) { outl (value, address); }
fnatr unsigned long inpd (unsigned long address) {return (inl (address));}
fnatr void outpd (unsigned long address, unsigned long value) { outl (address, value); }

fnatr void ModemRegOut8(unsigned int port, unsigned char databyte)  { outb ( databyte, (port + BaseAddress) ); }
fnatr void ModemRegOut16(unsigned int port, unsigned int dataword)  { outw( dataword, (port + BaseAddress)); }
fnatr void ModemRegOut32(unsigned int port, unsigned long dataword )	{ outl( dataword, (port + BaseAddress)); }
fnatr unsigned char ModemRegIn8(unsigned int port)  { return (inb (port + BaseAddress)); }
fnatr unsigned short ModemRegIn16(unsigned int port) { return (inw( port + BaseAddress)); }
fnatr unsigned long ModemRegIn32(unsigned int port) { return (inl( port + BaseAddress)); }

#ifndef LT_KER_26
void W_DOWN_IO (void) { __SLOW_DOWN_IO; }
#else
fnatr void x__SLOW_DOWN_IO (void) {}
#endif

/******************* Processor Flag Operations *********************/
#ifndef LT_KER_26
void x_save_flags (unsigned long *flags) { save_flags (*flags); }
void x_restore_flags (unsigned long *flags) { restore_flags (*flags); }
void x_sti (void){ sti (); }
void x_cli (void) { cli (); }
#else
fnatr void x_save_flags (unsigned long *flags) { spin_lock_irqsave(&modem_driver_lock, (*flags)); }
fnatr void x_restore_flags (unsigned long *flags) { spin_unlock_irqrestore(&modem_driver_lock, (*flags)); }
fnatr void x_sti (void){}
fnatr void x_cli (void) {}
#endif

/************************** IO - Interrupts ************************/
fnatr void *x_request_region(unsigned long start, unsigned long n, const char *name)
{
//  return(request_region(start, n, name));
  return(__request_region(&ioport_resource, (start), (n), (name)));
}

#ifndef LT_KER_26
int x_check_region(unsigned long start, unsigned long n) { return (check_region (start, n)); }
void x_release_region(unsigned long start, unsigned long n) { release_region(start, n); }
#else
fnatr int x_check_region(unsigned long start, unsigned long n) { return (__check_region(&ioport_resource, start, n)); }
fnatr void x_release_region(unsigned long start, unsigned long n) { __release_region(&ioport_resource, (start), (n)); }
#endif

fnatr int x_request_irq(unsigned int irq, void (*handler), unsigned long irqflags, const char * devname, void *dev_id)
{
  return (request_irq(irq, wrap_LX_isr_handler, irqflags, devname, dev_id));
}
fnatr void x_free_irq(unsigned int irq, void *dev_id) { free_irq( irq, dev_id); }

/********************** Time Operations *************************/
fnatr void x_ms_delay (unsigned ms) { mdelay (ms); }
fnatr void *x_do_gettimeofday(void *tval) { do_gettimeofday ((struct timeval *)tval); return(tval); }
fnatr void lt_add_timer( void (*timerfunction)(unsigned long) )
{
	timerList.expires = jiffies+1;
	timerList.function = wrap_timertick_function;
	timerList.data = 0;
	add_timer(&timerList);
}
fnatr dword VMODEM_Get_System_Time (void)
{
	struct timeval time;
	do_gettimeofday( &time);
	return((time.tv_usec/1000)+(time.tv_sec*1000));
}
fnatr void lt_init_timer ( void  ) { init_timer(&timerList); }

/******************** String / Mem Operations **********************/
fnatr char *x_strcpy (char *dest, const char *src) { return (strcpy (dest, src)); }
fnatr char *x_strncpy (char *dest, const char *src, int maxlen) { return (strncpy (dest, src, maxlen)); }
fnatr char *x_strcat (char *dest, const char *src) { return (strcat (dest, src)); }
fnatr int x_strlen (const char *str) { return (strlen (str)); }
fnatr char *x_strchr (const char *str, char c) { return (strchr (str, c)); }
fnatr void *x_memset (void *dest, char c, int len) { return (memset (dest, c, len)); }
fnatr void *x_memcpy (void *dest, const void *src, int len) { return (memcpy (dest, src, len)); }
fnatr int x_memcmp (const void *ptr1, const void *ptr2, int len) { return (memcmp (ptr1, ptr2, len)); }
fnatr unsigned char x_isprint (char c) { return(isprint (c)); }
fnatr unsigned char x_toupper (char c) { return toupper (c); }
fnatr int x_sprintf(char * buf, const char * fmt, ...)
{
	va_list args;
	int i;

	va_start(args, fmt);
	i=vsprintf(buf,fmt,args);
	va_end(args);
	return (i);
}

/************************** Debug Prints ***************************/
fnatr void linux_debug_message ( byte *msg ) { printk("<1>ltmodem: %s\n",msg); }

fnatr void x_linux_dbg_print(const char *fmt, ...)
	{
	va_list args;
	char buf[1536];

	va_start (args, fmt);
	vsprintf (buf, fmt, args);
	va_end (args);
	printk ("%s", buf);
	}

fnatr void x_linux_dbg_print_crit(const char *fmt, ...)
	{
	va_list args;
	char buf[1536];

	va_start (args, fmt);
	vsprintf (buf, fmt, args);
	va_end (args);
	printk (KERN_CRIT "%s", buf);
	}

/************************** PCI Operations *************************/
#ifndef LT_KER_26
int x_pcibios_present(void) { return(pcibios_present()); }
#else
fnatr int x_pcibios_present(void) { return(1); }
#endif

fnatr unsigned int agr_pci_find_device ( unsigned int vendor, unsigned int device, unsigned int *irqno )
{
//  static struct pci_dev *dev = NULL;
	struct pci_dev *dev = NULL;
 	dev = pci_find_device ( vendor, device, dev );
	if ((dev) && (dev->irq))
	{
		*irqno = dev->irq;
    return 1;
  }
  else
    return 0;
}
	
fnatr int x_pcibios_read_config_byte (unsigned char bus, unsigned char dev_fn, unsigned char where, unsigned char *val)
{struct pci_dev *dev = pci_find_slot(bus, dev_fn); return ((!dev)? PCIBIOS_DEVICE_NOT_FOUND : pci_read_config_byte(dev, where, val));}
            
fnatr int x_pcibios_read_config_word (unsigned char bus, unsigned char dev_fn, unsigned char where, unsigned short *val)
{struct pci_dev *dev = pci_find_slot(bus, dev_fn); return ((!dev)? PCIBIOS_DEVICE_NOT_FOUND : pci_read_config_word(dev, where, val));}

fnatr int x_pcibios_read_config_dword (unsigned char bus, unsigned char dev_fn, unsigned char where, unsigned int *val)
{struct pci_dev *dev = pci_find_slot(bus, dev_fn); return ((!dev)? PCIBIOS_DEVICE_NOT_FOUND : pci_read_config_dword(dev, where, val));}

fnatr int x_pcibios_write_config_byte (unsigned char bus, unsigned char dev_fn, unsigned char where, unsigned char val)
{struct pci_dev *dev = pci_find_slot(bus, dev_fn); return ((!dev)? PCIBIOS_DEVICE_NOT_FOUND : pci_write_config_byte(dev, where, val));}

fnatr int x_pcibios_write_config_word (unsigned char bus, unsigned char dev_fn, unsigned char where, unsigned short val)
{struct pci_dev *dev = pci_find_slot(bus, dev_fn); return ((!dev)? PCIBIOS_DEVICE_NOT_FOUND : pci_write_config_word(dev, where, val));}

fnatr int x_pcibios_write_config_dword (unsigned char bus, unsigned char dev_fn, unsigned char where, unsigned int val)
{struct pci_dev *dev = pci_find_slot(bus, dev_fn); return ((!dev)? PCIBIOS_DEVICE_NOT_FOUND : pci_write_config_dword(dev, where, val));}

#ifndef LT_KER_26
int x_pcibios_find_class (unsigned int class_code, unsigned short index, unsigned char *bus, unsigned char *dev_fn)
{
  return(pcibios_find_class (class_code, index, bus, dev_fn));
}
int x_pcibios_find_device (unsigned short vendor, unsigned short dev_id,
			 unsigned short index, unsigned char *bus,
			 unsigned char *dev_fn)
{
return(pcibios_find_device (vendor, dev_id, index, bus, dev_fn));
}
#endif

/************************** OS Operations **************************/
fnatr int x_smp_processor_id (void) { return (smp_processor_id ()); }
#ifndef LT_KER_26
int x_smp_num_cpus (void) {if (smp_num_cpus > 0) return smp_num_cpus;else return 1; }
#else
fnatr int x_smp_num_cpus (void) {if (num_online_cpus() > 0) return (num_online_cpus());else return 1; }
#endif

#ifndef LT_KER_26
void x_MOD_INC_USE_COUNT (void) { MOD_INC_USE_COUNT; }
void x_MOD_DEC_USE_COUNT (void) { MOD_DEC_USE_COUNT; }
#else
fnatr void x_MOD_INC_USE_COUNT (void) { try_module_get(THIS_MODULE); }
fnatr void x_MOD_DEC_USE_COUNT (void) { module_put(THIS_MODULE); }
#endif

fnatr void x_task_queue_init (x_tq_struct_t *task_x, x_func_ptr func, void *data)
	{
#ifndef LT_KER_26
	struct tq_struct *x_tqueue = (struct tq_struct *) task_x;
	INIT_TQUEUE(x_tqueue, func, data);
#else
	struct work_struct *x_tqueue = (struct work_struct *) task_x;
	INIT_WORK(x_tqueue, wrap_LXHardwareBottomHalf, data);
#endif
	}
fnatr void x_queue_task	(x_tq_struct_t *task_x)
	{
#ifndef LT_KER_26
	struct tq_struct *x_tqueue = (struct tq_struct *) task_x;
	queue_task (x_tqueue, &tq_immediate);
	mark_bh (IMMEDIATE_BH);
#else
	struct work_struct *x_tqueue = (struct work_struct *) task_x;
  schedule_work(x_tqueue);
#endif
	}

/************************ Atomic  Operations ***********************/
fnatr int x_atomic_read (int *atomic_x) { return atomic_read ((atomic_t *)&atomic_x); }
fnatr void x_atomic_set (int *atomic_x, int value) { atomic_set ((atomic_t *)&atomic_x, value); }
fnatr void x_atomic_inc (int *atomic_x) { atomic_inc ((atomic_t *)&atomic_x); }
fnatr void x_atomic_dec (int *atomic_x) {atomic_dec ((atomic_t *)&atomic_x); }
fnatr int x_atomic_dec_and_test (int *atomic_x) { return atomic_dec_and_test ((atomic_t *)&atomic_x)? 1: 0; }
